module Util where

import Control.Monad.Writer
import Data.Map (Map)
import qualified Data.Map as Map
import Debug.Trace

fromAllWriter :: Writer All a -> (Bool, a) 
fromAllWriter m = let (a, w) = runWriter m
                  in  (getAll w, a)

traceM :: (Monad m) => String -> m ()
traceM msg = trace msg $ return ()

liftMaybe :: (Monad m) => Maybe a -> m a
liftMaybe = maybe (fail "unexpected Nothing") return

(+?+) :: String -> String -> String
(+?+) xs "[]" = []
(+?+) xs ys = xs ++ ys 


mapMapM :: (Monad m, Ord k) => (a -> m b) -> Map k a -> m (Map k b)
mapMapM f = Map.foldWithKey step (return $ Map.empty)
  where step k a m = do a' <- f a
                        m' <- m
                        return $ Map.insert k a' m'

allM  :: Monad m => [m Bool] -> m Bool
allM [] = return True
allM (m:ms) = do b <- m
                 if b then allM ms else return False

findM :: Monad m => (a -> m Bool) -> [a] -> m (Maybe a)
findM p []       = return Nothing
findM p (x : xs) = do b <- p x 
                      if b then return (Just x) else findM p xs 
                        
parens :: String -> String
parens s = "(" ++ s ++ ")"

brackets :: String -> String 
brackets s = "[" ++ s ++ "]"

prettyList :: String -> (a -> String) -> [a] -> String
prettyList sep f [] = ""
prettyList sep f [e] = f e
prettyList sep f (e:es) = f e ++ sep ++ prettyList sep f es 

compressMaybes :: [Maybe a] -> [a]
compressMaybes = concat . map (maybe [] (\ a -> [a]))