{-# LANGUAGE TypeSynonymInstances #-}
 
module Value where

import Util

import Abstract

import Debug.Trace



-- call-by-value
-- cofuns are not forced
data Val =   VSet
           | VSize
           | VInfty
           | VSucc Val   
           | VApp Val [Val]
           | VCon Co Name
           | VDef Name               -- data/codata, fun/cofun identifier
           | VGen Int                -- free variable (de Bruijn level)
           | VCase Val Env [Clause]
           | VPi Dec Name Val Env Expr
           | VLam Name Env Expr
           | VMeta MVar Int          -- Xi + n  (n-fold successor of Xi)
           | VUp Val TVal            -- delayed eta expansion; TVal is a VPi
           | VIrr                    -- erased hypothetical inhabitant of empty type
             deriving (Eq)

type TVal = Val

arrow :: TVal -> TVal -> TVal
arrow a b = VPi defaultDec "" a [(bla,b)] (Var bla)
  where bla = "#ARROW#"

-- $ # = #
succSize :: Val -> Val
succSize v = case v of
            VInfty -> VInfty
            VMeta i n -> VMeta i (n + 1)
            _ -> VSucc v               


instance Show Val where
    show = showVal

showVal :: Val -> String
showVal VSet = "Set"
showVal VSize = "Size"
showVal VInfty = "#"
showVal (VSucc v) = "($ " ++ showVal v ++ ")" 
showVal (VApp v vl) = "(" ++ showVal v ++ " " ++ showVals vl ++ ")"
    where
      showVals :: [Val] -> String
      showVals [] = ""
      showVals (v:vl) = showVal v ++ (if null vl then "" else " " ++ showVals vl)   
showVal (VCon _ n) = n
showVal (VDef n) = n
showVal (VGen k) = "v" ++ show k
showVal (VMeta k 0) = "?" ++ show k
showVal (VMeta k 1) = "$?" ++ show k
showVal (VMeta k n) = "(?" ++ show k ++ "+" ++ show n ++")"
showVal (VCase v env cs) = "case " ++ showVal v ++ "{ " ++ prettyCases cs ++ " } " ++ showEnv env 
showVal (VLam x env e) = "(\\" ++ x ++ " -> " ++ show e ++ showEnv env ++ ")" 
showVal (VUp v vt) = "(" ++ show v ++ " Up " ++ show vt ++ ")" 
showVal VIrr  = "*"

showVal (VPi dec x av env b) =
  parens $ (if erased dec then brackets binding
              else if x=="" then s1 else parens binding) 
                                  ++ " -> " ++ show b ++ showEnv env
         where s1 = showVal av
               binding = if x == "" then s1 else x ++ " : " ++ s1 

{-
showVal (VPi dec "" av env e) = "(" ++ showVal av ++ " -> " ++ show e ++ showEnv env ++ ")"
showVal (VPi dec x av env e) = "((" ++ x  ++ " : " ++ showVal av ++ ") -> " ++ show e ++ showEnv env ++ ")"
-}

-- environment ---------------------------------------------------

type Env = [(Name,Val)]

emptyEnv = []

-- enviroment extension / update
update :: Env -> Name -> Val -> Env
update env "" v = env
update env n v = (n,v):env

lookupEnv :: Monad m => Env -> Name -> m Val
lookupEnv [] n = fail $ "lookup env error " ++ n 
lookupEnv ((x,v):xs) n = if x == n then return v 
                          else lookupEnv xs n

showEnv :: Env -> String
showEnv [] = ""
showEnv x = "{" ++ showEnv' x ++ "}" where
    showEnv' ((n,v):env) = n ++ " = " ++ showVal v ++ 
      (if null env then "" else ", " ++ showEnv' env) 

